import numpy as np
from scipy import stats
from joblib import Parallel, delayed
try:
    from dyfunconn import fc
except ImportError:
    fc = None


def _get_fc_graph(x, band_freq, sampling_freq, fc_measure='corr',
                  link_cutoff=0., band_freq_hi=(20., 45.), nfft=128,
                  n_overlap=64):
    """
    Build a functional connectivity network from the given data stream.
    :param x: numpy array of shape (n_channels, n_samples);
    :param band_freq: list with two elements, the band in which to estimate FC;
    :param sampling_freq: float, sampling frequency of the stream;
    :param fc_measure: functional connectivity measure to use;
    :param link_cutoff: links with absolute FC measure below this value will be
      removed;
    :param band_freq_hi: high band used to estimate FC when using 'aec';
    :param nfft: TODO, affects 'wpli' and 'dwpli';
    :param n_overlap: TODO, affects 'wpli' and 'dwpli';
    :return: FC graph in numpy format (note that node features are all ones).
    """
    if fc_measure == 'iplv':
        _, ef = fc.iplv(x, band_freq, sampling_freq)
    elif fc_measure == 'icoh':
        ef = fc.icoherence(x, band_freq, sampling_freq)
    elif fc_measure == 'corr':
        ef = fc.corr(x, band_freq, sampling_freq)
    elif fc_measure == 'aec':
        ef = fc.aec(x, band_freq, band_freq_hi, sampling_freq)
    elif fc_measure == 'wpli':
        csdparams = {'NFFT': nfft, 'noverlap': n_overlap}
        ef = fc.wpli(x, band_freq, sampling_freq, **csdparams)
        ef = np.abs(ef)
    elif fc_measure == 'dwpli':
        csdparams = {'NFFT': nfft, 'noverlap': n_overlap}
        ef = fc.dwpli(x, band_freq, sampling_freq, **csdparams)
        ef = np.abs(ef)
    elif fc_measure == 'dpli':
        ef = fc.dpli(x, band_freq, sampling_freq)
    else:
        raise ValueError('Invalid fc_measure')

    # Set the main diagonal to zero (no self-loops)
    np.fill_diagonal(ef, 0.0)

    # Compute adjacency matrix by rounding to 0 and 1 based on the cutoff
    adj = ef.copy()
    if link_cutoff != 0:
        adj[np.abs(adj) >= link_cutoff] = 1
        adj[np.abs(adj) < link_cutoff] = 0
    else:
        adj[...] = 1

    # Dummy node features
    # TODO: proper nf
    nf = np.ones((ef.shape[0], 1))

    # Edge features
    ef = ef[..., None]

    return adj, nf, ef


def get_fc(x, band_freq, sampling_freq, samples_per_graph=None,
           fc_measure='corr', link_cutoff=0., percentiles=None,
           band_freq_hi=(20., 45.), nfft=128, n_overlap=64, njobs=1):
    """
    Build functional connectivity networks from the given data stream.
    :param x: numpy array of shape (n_channels, n_samples);
    :param band_freq: list with two elements, the band in which to estimate FC;
    :param sampling_freq: float, sampling frequency of the stream;
    :param samples_per_graph: number of samples to use to generate a graph. By 
    default, the whole stream is used. If provided, 
    1 + (n_samples / samples_per_graph) will be generated;
    :param fc_measure: functional connectivity measure to use;
    :param link_cutoff: links with absolute FC measure below this value will be
      removed;
    :param percentiles: tuple of two numbers >0 and <100; links with FC measure 
    between the two percentiles will be removed (statistics are calculated for
    each edge). Note that this option ignores `link_cutoff`.
    :param band_freq_hi: high band used to estimate FC when using 'aec';
    :param nfft: TODO, affects 'wpli' and 'dwpli';
    :param n_overlap: TODO, affects 'wpli' and 'dwpli';
    :param njobs: number of processes to use (-1 to use all available cores);
    :return: FC graph(s) in numpy format (note that node features are all ones).
    """
    if fc is None:
        raise ImportError('`get_fc` requires dyfunconn.')
    if x.ndim != 2:
        raise ValueError('Expected x to have rank 2, got {}'.format(x.ndim))
    if samples_per_graph is None:
        samples_per_graph = x.shape[1]

    output = Parallel(n_jobs=njobs)(
        delayed(_get_fc_graph)(x[:, i: i + samples_per_graph],
                               band_freq,
                               sampling_freq,
                               fc_measure=fc_measure,
                               link_cutoff=link_cutoff,
                               band_freq_hi=band_freq_hi,
                               nfft=nfft,
                               n_overlap=n_overlap)
        for i in range(0, x.shape[1], samples_per_graph)
    )

    adj, nf, ef = zip(*output)
    adj = np.array(adj)
    nf = np.array(nf)
    ef = np.array(ef)

    if percentiles is not None:
        N = adj.shape[-1]
        pctl_lo = []
        pctl_hi = []
        for fcm in ef.reshape((-1, N * N)).T:
            pctl_lo.append(stats.scoreatpercentile(fcm, percentiles[0]))
            pctl_hi.append(stats.scoreatpercentile(fcm, percentiles[1]))
        pctl_lo = np.array(pctl_lo).reshape((1, N, N, 1))
        pctl_hi = np.array(pctl_hi).reshape((1, N, N, 1))
        adj = np.logical_or(ef < pctl_lo, ef > pctl_hi).astype(np.int)[..., 0]

    return adj, nf, ef



import numpy as np
try:
    from cdg.geometry import SphericalManifold, HyperbolicManifold
    CDG_OK = True
except ImportError:
    CDG_OK = False


# Euclidean manifold ###########################################################
def euclidean_distance(x, y):
    """
    Euclidean distance between points. Can be used as user-defined metric for
    sklearn.neighbors.DistanceMetric.
    :param x: one-dimensional np.array;
    :param y: one-dimensional np.array;
    :return: distance between the given points.
    """
    return np.linalg.norm(x - y, axis=-1, keepdims=True)


# Spherical manifold ###########################################################
def is_spherical(x, r=1.):
    """
    Boolean membership to spherical manifold.
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param r: positive float, the radius of the CCM;
    :return: boolean np.array, True if the points are on the CCM.
    """
    return (x ** 2).sum(-1).astype(np.float32) == r ** 2


def spherical_clip(x, r=1.):
    """
    Clips points in the ambient space to a spherical CCM of radius `r`.
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param r: positive float, the radius of the CCM;
    :return: np.array of same shape as x.
    """
    x = x.copy()
    norm = np.linalg.norm(x, axis=-1, keepdims=True)
    return r * (x / norm)


def spherical_distance(x, y):
    """
    Geodesic distance between points on a spherical CCM. Can be used as
    user-defined metric for sklearn.neighbors.DistanceMetric.
    :param x: one-dimensional np.array;
    :param y: one-dimensional np.array;
    :return: distance between the given points.
    """
    return np.arccos(np.clip(np.dot(x, y.T), -1, 1))


# Hyperbolic manifold ##########################################################
def is_hyperbolic(x, r=-1.):
    """
    Boolean membership to hyperbolic manifold.
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param r: negative float, the radius of the CCM;
    :return: boolean np.array, True if the points are on the CCM.
    """
    return ((x[..., :-1] ** 2).sum(-1) - x[..., -1] ** 2).astype(np.float32) == - r ** 2


def hyperbolic_clip(x, r=-1., axis=-1):
    """
    Clips points in the ambient space to a hyperbolic CCM of radius `r`, by f
    orcing the `axis` coordinate of the points to be
    \(X_{axis} = \sqrt{\sum\limits_{i \neq {axis}} X_{i}^{2} + r^{2}}\).
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param r: negative float, the radius of the CCM;
    :param axis: int, the axis along which to clip;
    :return: np.array of same shape as x.
    """
    x = x.copy()
    free_components_idxs = np.delete(np.arange(x.shape[-1]), axis)
    x[..., axis] = np.sqrt(np.sum(x[..., free_components_idxs] ** 2, -1) + (r ** 2))
    return x


def hyperbolic_inner(x, y):
    """
    Computes the inner product between points in the pseudo-euclidean
    ambient space of a hyperbolic manifold.
    Works also for 2D arrays of points.
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param y: np.array, coordinates are assumed to be in the last axis;
    :return: the inner product matrix.
    """
    minkowski_ipm = np.eye(x.shape[-1])
    minkowski_ipm[-1, -1] = -1
    inner = x.dot(minkowski_ipm).dot(y.T)
    return np.clip(inner, -np.inf, -1)


def hyperbolic_distance(x, y):
    """
    Geodesic distance between points on a hyperbolic CCM. Can be used as
    user-defined metric for sklearn.neighbors.DistanceMetric.
    :param x: one-dimensional np.array;
    :param y: one-dimensional np.array;
    :return: the computed distance.
    """
    inner = hyperbolic_inner(x, y)
    return np.arccosh(-inner)


# Generic CCM ##################################################################
def exp_map(x, r, tangent_point=None):
    """
    Let \(\mathcal{M}\) be a CCM of radius `r`, and \(T_{p}\mathcal{M}\) the
    tangent plane of the CCM at point \(p\) (`tangent_point`).
    This function maps a point `x` on the tangent plane to the CCM, using the
    Riemannian exponential map.
    :param x: np.array, point on the tangent plane (intrinsic coordinates);
    :param r: float, radius of the CCM;
    :param tangent_point: np.array, origin of the tangent plane on the CCM
    (extrinsic coordinates); if `None`, defaults to `[0., ..., 0., r]`.
    :return: the exp-map of x to the CCM (extrinsic coordinates).
    """
    if not CDG_OK:
        raise ImportError('`exp_map` requires CDG.')
    extrinsic_dim = x.shape[-1] + 1
    if tangent_point is None:
        tangent_point = np.zeros((extrinsic_dim,))
        tangent_point[-1] = np.abs(r)
    if isinstance(tangent_point, np.ndarray):
        if tangent_point.shape != (extrinsic_dim,) and tangent_point.shape != (1, extrinsic_dim):
            raise ValueError('Expected tangent_point of shape ({0},) or (1, {0}), got {1}'.format(extrinsic_dim, tangent_point.shape))
        if tangent_point.ndim == 1:
            tangent_point = tangent_point[np.newaxis, ...]
        if not belongs(tangent_point, r)[0]:
            raise ValueError('Tangent point must belong to manifold {}'.format(tangent_point))
    else:
        raise TypeError('tangent_point must be np.array or None')

    if r > 0.:
        return SphericalManifold.exp_map(tangent_point, x)
    elif r < 0.:
        return HyperbolicManifold.exp_map(tangent_point, x)
    else:
        return x


def log_map(x, r, tangent_point=None):
    """
    Let \(\mathcal{M}\) be a CCM of radius `r` and \(T_{p}\mathcal{M}\) the
    tangent plane of the CCM at point \(p\) (`tangent_point`).
    This function maps a point `x` on the CCM to the tangent plane, using the
    Riemannian logarithmic map.
    :param x: np.array, point on the CCM (extrinsic coordinates);
    :param r: float, radius of the CCM;
    :param tangent_point: np.array, origin of the tangent plane on the CCM
    (extrinsic coordinates); if 'None', defaults to `[0., ..., 0., r]`.
    :return: the log-map of x to the tangent plane (intrinsic coordinates).
    """
    if not CDG_OK:
        raise ImportError('`log_map` requires CDG.')
    extrinsic_dim = x.shape[-1]
    if tangent_point is None:
        tangent_point = np.zeros((extrinsic_dim,))
        tangent_point[-1] = np.abs(r)
    if isinstance(tangent_point, np.ndarray):
        if tangent_point.shape != (extrinsic_dim,) and tangent_point.shape != (1, extrinsic_dim):
            raise ValueError('Expected tangent_point of shape ({0},) or (1, {0}), got {1}'.format(extrinsic_dim, tangent_point.shape))
        if tangent_point.ndim == 1:
            tangent_point = tangent_point[np.newaxis, ...]
        if not belongs(tangent_point, r)[0]:
            raise ValueError('Tangent point must belong to manifold {}'.format(tangent_point))
    else:
        raise TypeError('tangent_point must be np.ndarray or None')

    if r > 0.:
        return SphericalManifold.log_map(tangent_point, x)
    elif r < 0.:
        return HyperbolicManifold.log_map(tangent_point, x)
    else:
        return x


def belongs(x, r):
    """
    Boolean membership to CCM of radius `r`.
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param r: float, the radius of the CCM;
    :return: boolean np.array, True if the points are on the CCM.
    """
    if r > 0.:
        return is_spherical(x, r)
    elif r < 0.:
        return is_hyperbolic(x, r)
    else:
        return np.ones(x.shape[:-1]).astype(np.float32)


def clip(x, r, axis=-1):
    """
    Clips points in the ambient space to a CCM of radius `r`.
    :param x: np.array, coordinates are assumed to be in the last axis;
    :param r: float, the radius of the CCM;
    :param axis: axis along which to clip points in the hyperbolic case (`r < 0`);
    :return: np.array of same shape as x.
    """
    if r > 0.:
        return spherical_clip(x, r)
    elif r < 0.:
        return hyperbolic_clip(x, r, axis=axis)
    else:
        return x


def get_distance(r):
    """
    :param r: float, the radius of the CCM;
    :return: the callable distance function for the CCM of radius `r`.
    """
    if r > 0.:
        return spherical_distance
    elif r < 0.:
        return hyperbolic_distance
    else:
        return euclidean_distance



from __future__ import absolute_import

import os

from keras.utils import get_file

from spektral.chem import sdf_to_nx
from spektral.utils.io import load_csv, load_sdf
from spektral.utils import nx_to_numpy

DATA_PATH = os.path.expanduser('~/.spektral/datasets/qm9/')
DATASET_URL = 'http://deepchem.io.s3-website-us-west-1.amazonaws.com/datasets/gdb9.tar.gz'
RETURN_TYPES = {'numpy', 'networkx', 'sdf'}
NODE_FEATURES = ['atomic_num', 'charge', 'coords', 'iso']
EDGE_FEATURES = ['type', 'stereo']
NF_PROCESSING = [None] * 6  # [label_to_one_hot_closure([6, 7, 8, 9])]
EF_PROCESSING = [None]      # [int_to_one_hot_closure(4)]
MAX_K = 9                   # Maximum number of nodes in a graph


def load_data(return_type='numpy', nf_keys=None, ef_keys=None, auto_pad=True,
              self_loops=False, amount=None):
    """
    Loads the QM9 molecules dataset.
    :param return_type: `'networkx'`, `'numpy'`, or `'sdf'`, data format to return;
    :param nf_keys: list or str, node features to return (see `qm9.NODE_FEATURES`
    for available features);
    :param ef_keys: list or str, edge features to return (see `qm9.NODE_FEATURES`
    for available features);
    :param auto_pad: if `return_type='numpy'`, zero pad graph matrices to have 
    the same dimension;
    :param self_loops: if `return_type='numpy'`, add self loops to adjacency 
    matrices;
    from the dataset). If `amount=None`, return 100 molecules by default;
    :param amount: the amount of molecules to return (works independently of 
    `debug`).
    :return: if `return_type='networkx'`, a list of graphs in Networkx format, 
    and a dataframe containing labels; if `return_type='numpy'`, the adjacency 
    matrix, node features, edge features, and a dataframe containing labels;
    if `return_type='sdf'`, a list of molecules in the internal SDF format and
    a dataframe containing labels.
    """
    if return_type not in RETURN_TYPES:
        raise ValueError('Possible return_type: {}'.format(RETURN_TYPES))

    if not os.path.exists(DATA_PATH):
        _ = dataset_downloader()  # Try to download dataset

    print('Loading QM9 dataset.')
    sdf_file = os.path.join(DATA_PATH, 'qm9.sdf')
    data = load_sdf(sdf_file, amount=amount)  # Internal SDF format

    # Load labels
    labels_file = os.path.join(DATA_PATH, 'qm9.sdf.csv')
    labels = load_csv(labels_file)
    if return_type is 'sdf':
        return data, labels
    else:
        # Convert to Networkx
        data = [sdf_to_nx(_) for _ in data]

    if return_type is 'numpy':
        if nf_keys is not None:
            if isinstance(nf_keys, str):
                nf_keys = [nf_keys]
        else:
            nf_keys = NODE_FEATURES
        if ef_keys is not None:
            if isinstance(ef_keys, str):
                ef_keys = [ef_keys]
        else:
            ef_keys = EDGE_FEATURES

        adj, nf, ef = nx_to_numpy(data,
                                  auto_pad=auto_pad, self_loops=self_loops,
                                  nf_keys=nf_keys, ef_keys=ef_keys,
                                  nf_postprocessing=NF_PROCESSING,
                                  ef_postprocessing=EF_PROCESSING)
        return adj, nf, ef, labels
    elif return_type is 'networkx':
        return data, labels
    else:
        # Should not get here
        raise RuntimeError()


def dataset_downloader():
    filename = get_file('qm9.tar.gz', DATASET_URL, extract=True,
                        cache_dir=DATA_PATH, cache_subdir=DATA_PATH)
    os.rename(DATA_PATH + 'gdb9.sdf', DATA_PATH + 'qm9.sdf')
    os.rename(DATA_PATH + 'gdb9.sdf.csv', DATA_PATH + 'qm9.sdf.csv')
    os.remove(DATA_PATH + 'qm9.tar.gz')
    return filename



import os

INIT_DISPLAY_OK = False
DISPLAY_AVAILABLE = True


def hide_gpu():
    """
    Hides visible CUDA devices so that computations will be run on the CPU only.
    """
    print('Hiding CUDA devices.')
    os.environ['CUDA_VISIBLE_DEVICES'] = ''


def n_gpus():
    """
    Returns the number of available GPUs in the system (requires Tensorflow).
    """
    try:
        from tensorflow.python.client import device_lib
        local_device_protos = device_lib.list_local_devices()
        return len([x.name for x in local_device_protos if x.device_type == 'GPU'])
    except ImportError:
        raise ImportError('`n_gpus` requires tensorflow')


def init_display():
    """
    Initializes Matplotlib's backend based on the available display.
    """
    global INIT_DISPLAY_OK, DISPLAY_AVAILABLE
    if 'DISPLAY' not in os.environ or os.environ['DISPLAY'].startswith('localhost'):
        import matplotlib
        matplotlib.use('Agg')
        DISPLAY_AVAILABLE = False
    INIT_DISPLAY_OK = True



from keras import backend as K
from keras.engine import Layer
from keras import regularizers, constraints, initializers


class GlobalAttentionPool(Layer):
    """
    A gated attention global pooling layer as presented by
    [Li et al. (2017)](https://arxiv.org/abs/1511.05493).
    Note that this layer assumes the `'channels_last'` data format, and cannot
    be used otherwise.

    **Mode**: single, batch.

    **Input**

    - node features of shape `(batch, num_nodes, num_features)`, depending on
    the mode;

    **Output**

    - a pooled feature matrix of shape `(batch, channels)`;

    **Arguments**

    - `channels`: integer, number of output channels;
    - `kernel_regularizer`: regularization applied to the gating networks;  

    **Usage**

    ```py
    X = Input(shape=(num_nodes, num_features))
    Z = GlobalAttentionPool(channels)(X)
    ```
        """
    def __init__(self, channels=32, kernel_regularizer=None, **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(GlobalAttentionPool, self).__init__(**kwargs)
        self.channels = channels
        self.kernel_regularizer = regularizers.get(kernel_regularizer)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        self.lg_kernel = self.add_weight('LG_kernel',
                                         (input_shape[-1], self.channels),
                                         initializer='glorot_uniform',
                                         regularizer=self.kernel_regularizer)
        self.lg_bias = self.add_weight('LG_bias',
                                       (self.channels,),
                                       initializer='zeros')
        self.attn_kernel = self.add_weight('attn_kernel',
                                           (input_shape[-1], self.channels),
                                           initializer='glorot_uniform',
                                           regularizer=self.kernel_regularizer)
        self.attn_bias = self.add_weight('attn_bias',
                                         (self.channels,),
                                         initializer='zeros')
        self.built = True

    def call(self, inputs):
        # Note that the layer assumes the 'channels_last' data format.
        inputs_linear = K.dot(inputs, self.lg_kernel) + self.lg_bias
        attn_map = K.dot(inputs, self.attn_kernel) + self.attn_bias
        attn_map = K.sigmoid(attn_map)
        masked_inputs = inputs_linear * attn_map
        output = K.sum(masked_inputs, 1)
        return output

    def compute_output_shape(self, input_shape):
        if len(input_shape) == 2:
            return input_shape[:-1] + (self.channels,)
        else:
            return (input_shape[0], self.channels)

    def get_config(self):
        config = {}
        base_config = super(GlobalAttentionPool, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


class NodeAttentionPool(Layer):
    """
    A node-attention global pooling layer. Pools a graph by learning attention
    coefficients to sum node features.
    Note that this layer assumes the `'channels_last'` data format, and cannot
    be used otherwise.

    **Mode**: single, batch.

    **Input**

    - node features of shape `(batch, num_nodes, num_features)`;

    **Output**

    - a pooled feature matrix of shape `(batch, num_features)`;

    **Arguments**

    - `attn_kernel_initializer`: initializer for the attention kernel matrix;
    - `kernel_regularizer`: regularization applied to the kernel matrix;  
    - `attn_kernel_regularizer`: regularization applied to the attention kernel 
    matrix;
    - `attn_kernel_constraint`: constraint applied to the attention kernel
    matrix;

    **Usage**
    ```py
    X = Input(shape=(num_nodes, num_features))
    Z = NodeAttentionPool()(X)
    ```
    """
    def __init__(self,
                 attn_kernel_initializer='glorot_uniform',
                 kernel_regularizer=None,
                 attn_kernel_regularizer=None,
                 attn_kernel_constraint=None,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)
        super(NodeAttentionPool, self).__init__(**kwargs)
        self.kernel_regularizer = regularizers.get(kernel_regularizer)
        self.attn_kernel_initializer = initializers.get(attn_kernel_initializer)
        self.attn_kernel_regularizer = regularizers.get(attn_kernel_regularizer)
        self.attn_kernel_constraint = constraints.get(attn_kernel_constraint)

    def build(self, input_shape):
        assert len(input_shape) >= 2
        # Attention kernels
        self.attn_kernel = self.add_weight(shape=(input_shape[-1], 1),
                                           initializer=self.attn_kernel_initializer,
                                           regularizer=self.attn_kernel_regularizer,
                                           constraint=self.attn_kernel_constraint,
                                           name='attn_kernel')

        self.built = True

    def call(self, inputs):
        input_shape = K.int_shape(inputs)
        # Note that the layer assumes the 'channels_last' data format.
        features = K.dot(inputs, self.attn_kernel)
        features = K.squeeze(features, -1)
        attn_coeff = K.softmax(features)  # TODO: maybe sigmoid?
        if len(input_shape) == 2:
            output = K.dot(attn_coeff, inputs)
        else:
            output = K.batch_dot(attn_coeff, inputs)

        return output

    def compute_output_shape(self, input_shape):
        if len(input_shape) == 2:
            return input_shape[-1:]
        else:
            return (input_shape[0], input_shape[-1])

    def get_config(self):
        config = {}
        base_config = super(NodeAttentionPool, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))




"""
The MIT License

Copyright (c) 2016 Thomas Kipf

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

This code was taken almost verbatim from https://github.com/tkipf/gcn/ and
adapted to work in Spektral.
"""
from __future__ import absolute_import

import os

import networkx as nx
import numpy as np
import scipy.sparse as sp

from spektral.utils.io import load_binary

DATA_PATH = os.path.expanduser('~/.spektral/datasets/')
AVAILABLE_DATASETS = {'cora', 'citeseer', 'pubmed'}
RETURN_TYPES = {'numpy'}


def _parse_index_file(filename):
    index = []
    for line in open(filename):
        index.append(int(line.strip()))
    return index


def _sample_mask(idx, l):
    mask = np.zeros(l)
    mask[idx] = 1
    return np.array(mask, dtype=np.bool)


def load_data(dataset_name='cora'):
    """
    Loads the specified citation dataset using the public splits as defined in
    [Kipf & Welling (2016)](https://arxiv.org/abs/1609.02907).
    :param dataset_name: name of the dataset to load (cora, citeseer, or pubmed)
    :return: the citation network in numpy format, with train, test, and
    validation splits for the targets and masks.
    """
    if dataset_name not in AVAILABLE_DATASETS:
        raise ValueError('Available datasets: {}'.format(AVAILABLE_DATASETS))

    if not os.path.exists(DATA_PATH + dataset_name):
        # TODO dataset downloader
        raise ValueError('Dataset not available. Download the data from '
                         'https://github.com/tkipf/gcn/tree/master/gcn/data and '
                         'place it in {}/'.format(DATA_PATH + dataset_name))

    print('Loading {} dataset'.format(dataset_name))

    names = ['x', 'y', 'tx', 'ty', 'allx', 'ally', 'graph']
    objects = []
    data_path = os.path.join(DATA_PATH, dataset_name)
    for n in names:
        filename = "{}/ind.{}.{}".format(data_path, dataset_name, n)
        objects.append(load_binary(filename))

    x, y, tx, ty, allx, ally, graph = tuple(objects)
    adj = nx.adjacency_matrix(nx.from_dict_of_lists(graph))
    test_idx_reorder = _parse_index_file("{}/ind.{}.test.index".format(data_path, dataset_name))
    test_idx_range = np.sort(test_idx_reorder)

    if dataset_name == 'citeseer':
        test_idx_range_full = range(min(test_idx_reorder),
                                    max(test_idx_reorder) + 1)
        tx_extended = sp.lil_matrix((len(test_idx_range_full), x.shape[1]))
        tx_extended[test_idx_range - min(test_idx_range), :] = tx
        tx = tx_extended
        ty_extended = np.zeros((len(test_idx_range_full), y.shape[1]))
        ty_extended[test_idx_range - min(test_idx_range), :] = ty
        ty = ty_extended

    features = sp.vstack((allx, tx)).tolil()
    features[test_idx_reorder, :] = features[test_idx_range, :]

    labels = np.vstack((ally, ty))
    labels[test_idx_reorder, :] = labels[test_idx_range, :]

    idx_test = test_idx_range.tolist()
    idx_train = range(len(y))
    idx_val = range(len(y), len(y) + val_size)

    train_mask = _sample_mask(idx_train, labels.shape[0])
    val_mask = _sample_mask(idx_val, labels.shape[0])
    test_mask = _sample_mask(idx_test, labels.shape[0])

    # Row-normalize the features
    # features_norm = utils.sparse_normalize(features)

    y_train = np.zeros(labels.shape)
    y_val = np.zeros(labels.shape)
    y_test = np.zeros(labels.shape)
    y_train[train_mask, :] = labels[train_mask, :]
    y_val[val_mask, :] = labels[val_mask, :]
    y_test[test_mask, :] = labels[test_mask, :]

    return adj, features, y_train, y_val, y_test, train_mask, val_mask, test_mask


def preprocess_features(features):
    rowsum = np.array(features.sum(1))
    r_inv = np.power(rowsum, -1).flatten()
    r_inv[np.isinf(r_inv)] = 0.
    r_mat_inv = sp.diags(r_inv)
    features = r_mat_inv.dot(features)
    return features



from __future__ import absolute_import

import numpy as np
from scipy.spatial import Delaunay

from spektral.utils import label_to_one_hot, numpy_to_nx

RETURN_TYPES = {'numpy', 'networkx'}
MAX_K = 7  # Maximum number of nodes in a graph


def generate_data(return_type='networkx', classes=0, n_samples_in_class=1000,
                  n_nodes=7, support_low=0., support_high=10., drift_amount=1.0,
                  one_hot_labels=True, support=None, seed=None):
    """
    Generates a dataset of Delaunay triangulations as described by
    [Zambon et al. (2017)](https://arxiv.org/abs/1706.06941).
    Note that this function is basically deprecated and will change soon.
    
    :param return_type: `'networkx'` or `'numpy'`, data format to return;
    :param classes: indices of the classes to load (integer, or list of integers
    between 0 and 20);
    :param n_samples_in_class: number of generated samples per class;
    :param n_nodes: number of nodes in a graph;
    :param support_low: lower bound of the uniform distribution from which the 
    support is generated;
    :param support_high: upper bound of the uniform distribution from which the 
    support is generated;
    :param drift_amount: coefficient to control the amount of change between 
    classes;
    :param one_hot_labels: one-hot encode dataset labels;
    :param support: custom support to use instead of generating it randomly; 
    :param seed: random numpy seed;
    :return: if `return_type='networkx'`, a list of graphs in Networkx format, 
    and an array containing labels; if `return_type='numpy'`, the adjacency 
    matrix, node features, and an array containing labels.
    """
    if return_type not in RETURN_TYPES:
        raise ValueError('Possible return_type: {}'.format(RETURN_TYPES))

    if isinstance(classes, int):
        classes = [classes]

    if max(classes) > 20 or min(classes) < 0:
        raise ValueError('Class indices must be between 0 and 20')

    r_classes = list(reversed(classes))
    if r_classes[-1] == 0:
        r_classes.insert(0, r_classes.pop(-1))

    # Support points
    np.random.seed(seed)
    if support is None:
        support = np.random.uniform(support_low, support_high, (1, n_nodes, 2))
    else:
        try:
            assert support.shape == (1, n_nodes, 2)
        except AssertionError:
            print('The given support doesn\'t have shape (1, n_nodes, 2) as'
                  'expected. Attempting to reshape.')
            support = support.reshape(1, n_nodes, 2)

    # Compute node features
    node_features = []
    # Other node features
    for idx, i in enumerate(r_classes):
        if i == 0:
            concept_0 = np.repeat(support, n_samples_in_class, 0)
            noise_0 = np.random.normal(0, 1, (n_samples_in_class, n_nodes, 2))
            class_0 = concept_0 + noise_0
            node_features.append(class_0)
        else:
            radius = 10. * ((2./3.) ** (drift_amount * (i - 1)))
            phase = np.random.uniform(0, 2 * np.pi, (n_nodes, 1))
            perturb_i_x = radius * np.cos(phase)
            perturb_i_y = radius * np.sin(phase)
            perturb_i = np.concatenate((perturb_i_x, perturb_i_y), axis=-1)
            support_i = support + perturb_i
            concept_i = np.repeat(support_i, n_samples_in_class, 0)
            noise_i = np.random.normal(0, 1, (n_samples_in_class, n_nodes, 2))
            class_i = concept_i + noise_i
            node_features.append(class_i)
    node_features = np.array(node_features).reshape((-1, n_nodes, 2))

    # Compute adjacency matrices
    adjacency = []
    for nf in node_features:
        adj = compute_adj(nf)
        adjacency.append(adj)
    adjacency = np.array(adjacency)

    # Compute labels
    labels = np.repeat(classes, n_samples_in_class)
    if one_hot_labels:
        labels = label_to_one_hot(labels, labels=classes)

    if return_type is 'numpy':
        return adjacency, node_features, labels
    elif return_type is 'networkx':
        graphs = numpy_to_nx(adjacency, node_features=node_features, nf_name='coords')
        return graphs, labels
    else:
        raise NotImplementedError


def compute_adj(x):
    """
    Computes the Delaunay triangulation of the given points
    :param x: array of shape (num_nodes, 2)
    :return: the computed adjacency matrix
    """
    tri = Delaunay(x)
    edges_explicit = np.concatenate((tri.vertices[:, :2],
                                     tri.vertices[:, 1:],
                                     tri.vertices[:, ::2]), axis=0)
    adj = np.zeros((x.shape[0], x.shape[0]))
    adj[edges_explicit[:, 0], edges_explicit[:, 1]] = 1.
    return np.clip(adj + adj.T, 0, 1)



import numpy as np
from spektral.geometric.manifold import spherical_clip, exp_map


# Uniform ######################################################################
def spherical_uniform(size, dim=3, r=1.):
    """
    Samples points from a uniform distribution on a spherical manifold.
    Uniform sampling on the sphere can be achieved by sampling from a Gaussian
    in the ambient space of the CCM, and then projecting the samples onto the
    sphere.
    :param size: number of points to sample;
    :param dim: dimension of the ambient space;
    :param r: positive float, the radius of the CCM;
    :return: np.array of shape (size, dim).
    """
    samples = np.random.normal(0, 1, (size, dim))
    samples = spherical_clip(samples, r=r)
    return samples


def hyperbolic_uniform(size, dim=3, r=-1., low=-1., high=1., projection='upper'):
    """
    Samples points from a uniform distribution on a hyperbolic manifold. Uniform
    sampling on a hyperbolic CCM can be achieved by sampling from a uniform
    distribution in the ambient space of the CCM, and then projecting the
    samples onto the CCM.
    :param size: number of points to sample;
    :param dim: dimension of the ambient space;
    :param r: negative float, the radius of the CCM;
    :param low: lower bound of the uniform distribution from which to sample;
    :param high: upper bound of the uniform distribution from which to sample;
    :param projection: 'upper', 'lower', or 'both'. Whether to project points 
    always on the upper or lower branch of the hyperboloid, or on both based 
    on the sign of the last coordinate. 
    :return: np.array of shape (size, dim).
    """
    samples = np.random.uniform(low, high, (size, dim))
    if projection == 'both':
        sign = np.sign(samples[..., -1:])
    elif projection == 'upper':
        sign = 1
    elif projection == 'lower':
        sign = -1
    else:
        raise NotImplementedError('Possible projection modes: \'both\', '
                                  '\'upper\', \'lower\'.')
    samples[..., -1:] = sign * np.sqrt((samples[..., :-1] ** 2).sum(-1, keepdims=True) + r ** 2)

    return samples


def _ccm_uniform(size, dim=3, r=0., low=-1., high=1., projection='upper'):
    """
    Samples points from a uniform distribution on a constant-curvature manifold.
    If `r=0`, then points are sampled from a uniform distribution in the ambient
    space.
    :param size: number of points to sample;
    :param dim: dimension of the ambient space;
    :param r: float, the radius of the CCM;
    :param low: lower bound of the uniform distribution from which to sample;
    :param high: upper bound of the uniform distribution from which to sample;
    :param projection: 'upper', 'lower', or 'both'. Whether to project points 
    always on the upper or lower branch of the hyperboloid, or on both based 
    on the sign of the last coordinate.
    :return: np.array of shape (size, dim).
    """
    if r < 0.:
        return hyperbolic_uniform(size, dim=dim, r=r, low=low, high=high,
                                  projection=projection)
    elif r > 0.:
        return spherical_uniform(size, dim=dim, r=r)
    else:
        return np.random.uniform(low, high, (size, dim))


def ccm_uniform(size, dim=3, r=0., low=-1., high=1., projection='upper'):
    """
    Samples points from a uniform distribution on a constant-curvature manifold.
    If `r=0`, then points are sampled from a uniform distribution in the ambient
    space.
    If a list of radii is passed instead of a single scalar, then the sampling
    is repeated for each value in the list and the results are concatenated
    along the last axis (e.g., see [Grattarola et al. (2018)](https://arxiv.org/abs/1805.06299)).
    :param size: number of points to sample;
    :param dim: dimension of the ambient space;
    :param r: floats or list of floats, radii of the CCMs;
    :param low: lower bound of the uniform distribution from which to sample;
    :param high: upper bound of the uniform distribution from which to sample;
    :param projection: 'upper', 'lower', or 'both'. Whether to project points
    always on the upper or lower branch of the hyperboloid, or on both based
    on the sign of the last coordinate.
    :return: if `r` is a scalar, np.array of shape (size, dim). If `r` is a
    list, np.array of shape (size, len(r) * dim).
    """
    if isinstance(r, int) or isinstance(r, float):
        r = [r]
    elif isinstance(r, list) or isinstance(r, tuple):
        r = r
    else:
        raise TypeError('Radius must be either a single value, a list'
                        'of values (or a tuple).')
    to_ret = []
    for r_ in r:
        to_ret.append(_ccm_uniform(size, dim=dim, r=r_, low=low, high=high,
                                   projection=projection))
    return np.concatenate(to_ret, -1)


# Normal #######################################################################
def spherical_normal(size, tangent_point, r, dim=3, loc=0., scale=1.):
    """
    Samples points from a normal distribution on a spherical manifold.
    Normal sampling on the sphere works by sampling from a Gaussian on the
    tangent plane, and then projecting the sampled points onto the sphere using
    the Riemannian exponential map.
    :param size: number of points to sample;
    :param tangent_point: np.array, origin of the tangent plane on the CCM
    (extrinsic coordinates);
    :param dim: dimension of the ambient space;
    :param r: positive float, the radius of the CCM;
    :param loc: mean of the Gaussian on the tangent plane;
    :param scale: standard deviation of the Gaussian on the tangent plane;
    :return: np.array of shape (size, dim).
    """
    samples = np.random.normal(loc=loc, scale=scale, size=(size, dim - 1))
    samples = exp_map(samples, r, tangent_point)
    return samples


def hyperbolic_normal(size, tangent_point, r, dim=3, loc=0., scale=1.):
    """
    Samples points from a normal distribution on a hyperbolic manifold.
    Normal sampling on a hyperbolic CCM works by sampling from a Gaussian on the
    tangent plane, and then projecting the sampled points onto the CCM using
    the Riemannian exponential map.
    :param size: number of points to sample;
    :param tangent_point: np.array, origin of the tangent plane on the CCM
    (extrinsic coordinates);
    :param r: positive float, the radius of the CCM;
    :param dim: dimension of the ambient space;
    :param loc: mean of the Gaussian on the tangent plane;
    :param scale: standard deviation of the Gaussian on the tangent plane;
    :return: np.array of shape (size, dim).
    """
    samples = np.random.normal(loc=loc, scale=scale, size=(size, dim - 1))
    return exp_map(samples, r, tangent_point)


def _ccm_normal(size, dim=3, r=0., tangent_point=None, loc=0., scale=1.):
    """
    Samples points from a Gaussian distribution on a constant-curvature manifold.
    If `r=0`, then points are sampled from a Gaussian distribution in the
    ambient space.
    :param size: number of points to sample;
    :param tangent_point: np.array, origin of the tangent plane on the CCM
    (extrinsic coordinates); if 'None', defaults to `[0., ..., 0., r]`.
    :param r: float, the radius of the CCM;
    :param dim: dimension of the ambient space;
    :param loc: mean of the Gaussian on the tangent plane;
    :param scale: standard deviation of the Gaussian on the tangent plane;
    :return: np.array of shape (size, dim).
    """
    if tangent_point is None:
        tangent_point = np.zeros((dim, ))
        tangent_point[-1] = np.abs(r)
    if r < 0.:
        return hyperbolic_normal(size, tangent_point, r, dim=dim, loc=loc, scale=scale)
    elif r > 0.:
        return spherical_normal(size, tangent_point, r, dim=dim, loc=loc, scale=scale)
    else:
        return np.random.normal(loc, scale, (size, dim))


def ccm_normal(size, dim=3, r=0., tangent_point=None, loc=0., scale=1.):
    """
    Samples points from a Gaussian distribution on a constant-curvature manifold.
    If `r=0`, then points are sampled from a Gaussian distribution in the
    ambient space.
    If a list of radii is passed instead of a single scalar, then the sampling
    is repeated for each value in the list and the results are concatenated
    along the last axis (e.g., see [Grattarola et al. (2018)](https://arxiv.org/abs/1805.06299)).
    :param size: number of points to sample;
    :param tangent_point: np.array, origin of the tangent plane on the CCM
    (extrinsic coordinates); if 'None', defaults to `[0., ..., 0., r]`.
    :param r: floats or list of floats, radii of the CCMs;
    :param dim: dimension of the ambient space;
    :param loc: mean of the Gaussian on the tangent plane;
    :param scale: standard deviation of the Gaussian on the tangent plane;
    :return: if `r` is a scalar, np.array of shape (size, dim). If `r` is a
    list, np.array of shape (size, len(r) * dim).
    """
    if isinstance(r, int) or isinstance(r, float):
        r = [r]
    elif isinstance(r, list) or isinstance(r, tuple):
        r = r
    else:
        raise TypeError('Radius must be either a single value, a list'
                        'of values (or a tuple).')

    if tangent_point is None:
        tangent_point = [None] * len(r)
    elif isinstance(tangent_point, np.ndarray):
        tangent_point = [tangent_point]
    elif isinstance(tangent_point, list) or isinstance(tangent_point, tuple):
        pass
    else:
        raise TypeError('tangent_point must be either a single point or a'
                        'list of points.')

    if len(r) != len(tangent_point):
        raise ValueError('r and tangent_point must have the same length')

    to_ret = []
    for r_, tp_ in zip(r, tangent_point):
        to_ret.append(_ccm_normal(size, dim=dim, r=r_, tangent_point=tp_,
                                  loc=loc, scale=scale))
    return np.concatenate(to_ret, -1)


# Generic ######################################################################
def get_ccm_distribution(name):
    """
    :param name: 'uniform' or 'normal', name of the distribution.
    :return: the callable function for sampling on a generic CCM;
    """
    if name == 'uniform':
        return ccm_uniform
    elif name == 'normal':
        return ccm_normal
    else:
        raise ValueError('Possible distributions: \'uniform\', \'normal\'')



import numpy as np
import scipy.sparse as sp
from keras.datasets import mnist as m
from scipy.spatial.distance import cdist, squareform, pdist
from sklearn.model_selection import train_test_split
from sklearn.neighbors import kneighbors_graph


def load_data():
    """
    Loads the MNIST dataset and the associated grid.
    This code is largely taken from [Michaël Defferrard's Github](https://github.com/mdeff/cnn_graph/blob/master/nips2016/mnist.ipynb).

    :return:
        - X_train, y_train: training node features and labels;
        - X_val, y_val: validation node features and labels;
        - X_test, y_test: test node features and labels;
        - A: adjacency matrix of the grid;
    """

    A = grid_graph(28, corners=False)
    A = replace_random_edges(A, 0).astype(np.float32)

    (X_train, y_train), (X_test, y_test) = m.load_data()
    X_train, X_test = X_train / 255.0, X_test / 255.0
    X_train = X_train.reshape(-1, 28 * 28)
    X_test = X_test.reshape(-1, 28 * 28)

    X_train, X_val, y_train, y_val = train_test_split(X_train, y_train, test_size=10000)

    return X_train, y_train, X_val, y_val, X_test, y_test, A


def replace_random_edges(A, noise_level):
    """Replace randomly chosen edges by random edges."""
    M, M = A.shape
    n = int(noise_level * A.nnz // 2)

    indices = np.random.permutation(A.nnz//2)[:n]
    rows = np.random.randint(0, M, n)
    cols = np.random.randint(0, M, n)
    vals = np.random.uniform(0, 1, n)
    assert len(indices) == len(rows) == len(cols) == len(vals)

    A_coo = sp.triu(A, format='coo')
    assert A_coo.nnz == A.nnz // 2
    assert A_coo.nnz >= n
    A = A.tolil()

    for idx, row, col, val in zip(indices, rows, cols, vals):
        old_row = A_coo.row[idx]
        old_col = A_coo.col[idx]

        A[old_row, old_col] = 0
        A[old_col, old_row] = 0
        A[row, col] = 1
        A[col, row] = 1

    A.setdiag(0)
    A = A.tocsr()
    A.eliminate_zeros()
    return A


def eval_bw(X, Y):
    """
    Compute heuristically the bandwidth using class information
    Returns (d^2)/9, with d minimum distance of elements in X with different class Y
    A small value is added to avoid returning bw=0
    """
    classes = np.unique(Y)
    min_dist = np.inf
    for i in range(classes.shape[0] - 1):
        c_i = classes[i]
        X_i = X[Y == c_i, :]
        for j in range(i + 1, classes.shape[0]):
            c_j = classes[j]
            X_j = X[Y == c_j, :]
            dist_ij = np.min(cdist(X_i, X_j, metric='sqeuclidean'))
            if dist_ij < min_dist:
                min_dist = dist_ij

    return min_dist / 9.0 + 1e-6


def _grid(m, dtype=np.float32):
    """Returns the embedding of a grid graph."""
    M = m**2
    x = np.linspace(0, 1, m, dtype=dtype)
    y = np.linspace(0, 1, m, dtype=dtype)
    xx, yy = np.meshgrid(x, y)
    z = np.empty((M, 2), dtype)
    z[:, 0] = xx.reshape(M)
    z[:, 1] = yy.reshape(M)
    return z


def get_adj_from_data(X_l, Y_l=None, X_u=None, adj='knn', k=10, knn_mode='distance', metric='euclidean',
                      self_conn=True):
    """
    :param X_l: labelled node features;
    :param Y_l: labels associated to X_l;
    :param X_u: unlabelled node features;
    :param adj: type of adjacency matrix to compute.
        - 'rbf' to compute rbf with bandwidth evaluated heuristically with
        eval_bw;
        - 'knn' to compute a kNN graph (k must be specified)
    :param k: number of neighbors in the kNN graph or in the linear neighborhood;
    :param knn_mode: 'connectivity' (graph with 0 and 1) or 'distance';
    :param metric: metric to use to build the knn graph (see sklearn.neighbors.kneighbors_graph)
    :param self_conn: if True, self connections are removed from adj matrix (A_ii = 0)
    :return: adjacency matrix as a sparse array (knn) or numpy array (rbf)
    """
    if adj not in {'rbf', 'knn'}:
        raise ValueError('adj must be either rbf or knn')
    if X_u is not None:
        X = np.concatenate((X_l, X_u), axis=0)
    else:
        X = X_l

    # Compute transition prob matrix
    if adj == 'rbf':
        # Estimate bandwidth
        if Y_l is None:
            bw = 0.01
        else:
            bw = eval_bw(X_l, np.argmax(Y_l, axis=1))

        # Compute adjacency matrix
        d = squareform(pdist(X, metric='sqeuclidean'))
        A = np.exp(-d / bw)

        # No self-connections (avoids self-reinforcement)
        if self_conn is False:
            np.fill_diagonal(A, 0.0)
    elif adj == 'knn':
        if k is None:
            raise ValueError('k must be specified when adj=\'knn\'')
        # Compute adjacency matrix
        A = kneighbors_graph(
            X, n_neighbors=k,
            mode=knn_mode,
            metric=metric,
            include_self=self_conn
        ).toarray()
        A = sp.csr_matrix(np.maximum(A, A.T))
    else:
        raise NotImplementedError()

    return A


def grid_graph(m, corners=False):
    z = _grid(m)
    A = get_adj_from_data(z, adj='knn', k=8, metric='euclidean')

    # Connections are only vertical or horizontal on the grid.
    # Corner vertices are connected to 2 neightbors only.
    if corners:
        A = A.toarray()
        A[A < A.max()/1.5] = 0
        A = sp.csr_matrix(A)
        print('{} edges'.format(A.nnz))

    return A